<'
# Copyright (c) 2012-2014, 2017-2020, 2023, 2025 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Mar 14 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include <stdio.h>

#include "h2devLib.h"
#include "posterLib.h"

#include "<"$comp">/c/client.h"
#include "<"$comp">_msglib.h"
#include "<"$comp">_portlib.h"
#include "private.h"
#include "serialize.h"


/* --- local data ---------------------------------------------------------- */

<'foreach s [$component services] {'>
int	genom_<"$comp">_<"[$s name]">_encode(char *buffer, int size,
		char *dst, int maxsize);
int	genom_<"$comp">_<"[$s name]">_decode(char *buffer, int size,
		char *dst, int maxsize);
<'}'>
<'foreach p [$component ports out] {'>
int	genom_<"$comp">_<"[$p name]">_decode(char *buffer,
		<"[[$p datatype] argument reference dst]">);
<'}'>


/* <"[--- ${comp}_abort -------------------------------------------------]"> */

static void
genom_<"$comp">_client_abort_cb(genom_client h, int rqstid, int done,
                       genom_event report, void *output, const void *exdetail,
                       void *cb_data)
{
  (void)done; (void)report; (void)output;
  (void)exdetail; (void)cb_data; /* fix -Wunused-parameter */

  (void)pocolibs_clean(h, rqstid);
}

genom_event
genom_<"$comp">_client_abort(genom_client h, int rqstid)
{
  int abrtid;

  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_no_such_activity(&h->context);
  if (h->active_rqst[rqstid].status == FREE ||
      h->active_rqst[rqstid].csstatus == FREE)
    return genom_no_such_activity(&h->context);
  if (h->active_rqst[rqstid].csstatus == WAITING_INTERMED_REPLY)
    return genom_syserr(&(genom_syserr_detail){.code = EAGAIN}, &h->context);

  return pocolibs_sendrqst(
    h, <"$COMP">_abort_activity_RQSTID,
    genom_<"$comp">_client_abort_activity_info,
    &h->active_rqst[rqstid].aid, genom_<"$comp">_abort_activity_encode,
    FALSE,
    (void (*)(void *))genom_<"$comp">_client_abort_activity_init_output,
    genom_<"$comp">_abort_activity_decode,
    sizeof(struct genom_<"$comp">_abort_activity_output),
    NULL, genom_<"$comp">_client_abort_cb, NULL,
    &abrtid);
}

<'foreach s [$component services] {'>

/* <"[--- ${comp}_[$s name]_rqst ----------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$s name]">_rqst(
  genom_client h,
  const struct genom_<"$comp">_<"[$s name]">_input *in,
  <"$comp">_<"[$s name]">_cb sentcb, <"$comp">_<"[$s name]">_cb donecb,
  void *cb_data,
  int *rqstid)
{
<'  set noack [catch {$s task}]'>
  assert(h);

  return pocolibs_sendrqst(
    h, <"$COMP">_<"[$s name]">_RQSTID,
    genom_<"$comp">_client_<"[$s name]">_info,
    in, genom_<"$comp">_<"[$s name]">_encode,
    <"[expr ${noack} ? "FALSE" : "TRUE"]">,
    (void (*)(void *))genom_<"$comp">_client_<"[$s name]">_init_output,
    genom_<"$comp">_<"[$s name]">_decode,
    sizeof(struct genom_<"$comp">_<"[$s name]">_output),
    (genom_request_cb)sentcb, (genom_request_cb)donecb, cb_data,
    rqstid);
}


/* <"[--- ${comp}_[$s name]_result --------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$s name]">_result(
  genom_client h, int rqstid, genom_event *report,
  struct genom_<"$comp">_<"[$s name]">_output **out, void **exdetail)
{
  struct genom_client_request *a;
  genom_event s;
  void *p;

  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  a = &h->active_rqst[rqstid];
  if (a->info != genom_<"$comp">_client_<"[$s name]">_info)
    return genom_syserr(&(genom_syserr_detail){.code = ENOMSG}, &h->context);

  s = pocolibs_result(h, rqstid, report, &p, exdetail);
  if (s) return s;
  if (out) *out = p;
  return genom_ok;
}

<'}'>

<'foreach p [$component ports out] {'>

/* <"[--- ${comp}_[$p name] reader --------------------------------------]"> */

<"[[$p datatype] argument reference]">
genom_<"$comp">_client_<"[$p name]">_data(
<'  if {"multiple" in [$p kind]} {'>
    genom_client h, const char *name
<'  } else {'>
    genom_client h
<'  }'>)
{
<'  switch -- [$p kind] {'>
<'    simple {'>
  return <"[[$p datatype] pass reference h->ports->[$p name].data]">;
<'    }'>
<'    multiple {'>
  unsigned int i;

  for(i = 0; i < h->ports-><"[$p name]">.n; i++)
    if (!strcmp(h->ports-><"[$p name]">.h[i].name, name)) break;
  if (i >= h->ports-><"[$p name]">.n) {
    genom_syserr(&(genom_syserr_detail){.code = ENXIO}, &h->context);
    return NULL;
  }
  return <"[[$p datatype] pass reference h->ports->[$p name].h\[i\].data]">;
<'    }'>
<'  }'>
}

genom_event
genom_<"$comp">_client_<"[$p name]">_read(
  genom_client h,
<'if {"multiple" in [$p kind]} {'>
  const char *path,
<'}'>
  <"[[$p datatype] argument reference data]">)
{
  POSTER_ID *id;
  int searched = 0;
  char *addr;
  int s;

  /* cached id */
<'  switch -- [$p kind] {'>
<'    simple {'>
  id = &h->ports-><"[$p name]">.id;
  if (!data)
    data = <"[[$p datatype] pass reference h->ports->[$p name].data]">;
<'    }'>
<'    multiple {'>
  unsigned int i;

  for(i = 0; i < h->ports-><"[$p name]">.n; i++) {
    if (!strcmp(h->ports-><"[$p name]">.h[i].name, path)) break;
    if (h->ports-><"[$p name]">.h[i].name[0] == 0) {
      /* reuse empty slot */
      snprintf(h->ports-><"[$p name]">.h[i].name,
               sizeof(h->ports-><"[$p name]">.h[i].name), "%s", path);
      genom_<"$comp">_client_<"[$p name]">_init_data(
        <"[[$p datatype] pass reference h->ports->[$p name].h\[i\].data]">);
      break;
    }
  }
  if (i >= h->ports-><"[$p name]">.n) {
    void *a = realloc(
      h->ports-><"[$p name]">.h, (i+1) * sizeof(*h->ports-><"[$p name]">.h));
    if (!a)
      return genom_syserr(
        &(genom_syserr_detail){.code = errno}, &h->context);

    h->ports-><"[$p name]">.h = a;
    snprintf(h->ports-><"[$p name]">.h[i].name,
             sizeof(h->ports-><"[$p name]">.h[i].name), "%s", path);
    genom_<"$comp">_client_<"[$p name]">_init_data(
      <"[[$p datatype] pass reference h->ports->[$p name].h\[i\].data]">);
    h->ports-><"[$p name]">.h[i].id = NULL;
    h->ports-><"[$p name]">.n = i+1;
  }

  id = &h->ports-><"[$p name]">.h[i].id;
  if (!data)
    data = <"[[$p datatype] pass reference h->ports->[$p name].h\[i\].data]">;
<'    }'>
<'  }'>

  do {
    if (!*id) {
      char name[H2_DEV_MAX_NAME];
      int n;
<'  switch -- [$p kind] {'>
<'    simple {'>
      n = snprintf(name, sizeof(name), "%s/<"[$p name]">", h->instance);
<'    }'>
<'    multiple {'>
      n = snprintf(
        name, sizeof(name), "%s/<"[$p name]">/%s", h->instance, path);
<'    }'>
<'  }'>
      if (n <= 0 || n >= (signed)sizeof(name)) {
<'  if {"multiple" in [$p kind]} {'>
        h->ports-><"[$p name]">.h[i].name[0] = 0;
        h->ports-><"[$p name]">.h[i].id = NULL;
<'  }'>
        return genom_syserr(
          &(genom_syserr_detail){.code = ENAMETOOLONG}, &h->context);
      }

      if (posterFind(name, id) != OK) {
        genom_mwerr_detail d;
        char msg[sizeof(d.what)];
        h2getErrMsg(errnoGet(), msg, sizeof(msg));
        if (snprintf(d.what, sizeof(d.what), "%s port: %s", name, msg) < 0)
          strncpy(d.what, "internal error", sizeof(d.what));
<'  if {"multiple" in [$p kind]} {'>
        h->ports-><"[$p name]">.h[i].name[0] = 0;
        h->ports-><"[$p name]">.h[i].id = NULL;
<'  }'>
        return genom_mwerr(&d, &h->context);
      }
      searched = 1;
    }

    if (posterTake(*id, POSTER_READ) != OK) {
      genom_mwerr_detail d;
      char msg[sizeof(d.what)];
      int code = errnoGet();
      if (!searched && code == S_posterLib_POSTER_CLOSED) {
        *id = NULL; continue /* try to refetch poster id */;
      }

      h2getErrMsg(code, msg, sizeof(msg));
<'  switch -- [$p kind] {'>
<'    simple {'>
      if (snprintf(d.what, sizeof(d.what),
                   "%s/<"[$p name]"> port: %s", h->instance, msg) < 0)
        strncpy(d.what, "internal error", sizeof(d.what));
<'    }'>
<'    multiple {'>
      if (snprintf(d.what, sizeof(d.what),
                   "%s/%s port: %s", h->instance, path, msg) < 0)
        strncpy(d.what, "internal error", sizeof(d.what));
<'    }'>
<'  }'>
      return genom_mwerr(&d, &h->context);
    }

    break;
  } while(1);

  addr = posterAddr(*id);
  s = genom_<"$comp">_<"[$p name]">_decode(addr, data);
  posterGive(*id);

  if (s)
    return genom_syserr(&(genom_syserr_detail){.code = s}, &h->context);
  return genom_ok;
}

genom_event
genom_<"$comp">_client_<"[$p name]">_port(
  genom_client h,
<'if {"multiple" in [$p kind]} {'>
  const char *path,
<'}'>
  <"[[$p datatype] argument reference data]">)
  __attribute((alias("genom_<"$comp">_client_<"[$p name]">_read")));

<'}'>
<'foreach p [$component ports in] {'>


/* <"[--- ${comp}_[$p name] publisher -----------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$p name]">_open(
  genom_client h, const char *path, const char **name)
{
  size_t l;
  int i, n;

  /* find slot */
  if (!*path)
    return genom_syserr(&(genom_syserr_detail){.code = ENXIO}, &h->context);

  for(i = 0; i < h->ports-><"[$p name]">.n; i++) {
    if (!strcmp(h->ports-><"[$p name]">.h[i].name, path)) break;
    if (h->ports-><"[$p name]">.h[i].name[0] == 0)
      /* reuse empty slot */ break;
  }
  if (i >= h->ports-><"[$p name]">.n) {
    /* create slot */
    void *a = realloc(
      h->ports-><"[$p name]">.h, (i+1) * sizeof(*h->ports-><"[$p name]">.h));
    if (!a)
      return genom_syserr(&(genom_syserr_detail){.code = errno}, &h->context);

    h->ports-><"[$p name]">.h = a;
    genom_<"$comp">_client_<"[$p name]">_init_data(
      <"[[$p datatype] pass reference h->ports->[$p name].h\[i\].data]">);
    h->ports-><"[$p name]">.h[i].id = NULL;
    h->ports-><"[$p name]">.n = i+1;
  }
  *name = h->ports-><"[$p name]">.h[i].name;

  if (h->ports-><"[$p name]">.h[i].id)
    /* already exists locally */ return genom_ok;

  /* set name */
  n = snprintf(h->ports-><"[$p name]">.h[i].name,
               sizeof(h->ports-><"[$p name]">.h[i].name), "%s", path);
  if (n <= 0 || n >= (signed)sizeof(h->ports-><"[$p name]">.h[i].name)) {
    h->ports-><"[$p name]">.h[i].name[0] = 0;
    return genom_syserr(
      &(genom_syserr_detail){.code = ENAMETOOLONG}, &h->context);
  }

  /* create */
  l = genom_maxserialen_<"[[$p datatype] mangle]">();
  if (posterCreate(*name, (int)l, &h->ports-><"[$p name]">.h[i].id) != OK) {
    genom_mwerr_detail d;
    char msg[sizeof(d.what)];

    h2getErrMsg(errnoGet(), msg, sizeof(msg));
    if (snprintf(d.what, sizeof(d.what), "%s port: %s", path, msg) < 0)
      strncpy(d.what, "internal error", sizeof(d.what));
    h->ports-><"[$p name]">.h[i].name[0] = 0;
    h->ports-><"[$p name]">.h[i].id = NULL;
    return genom_mwerr(&d, &h->context);
  }
  h->ports-><"[$p name]">.h[i].size = l;

  return genom_ok;
}

genom_event
genom_<"$comp">_client_<"[$p name]">_close(
  genom_client h, const char *name)
{
  POSTER_ID p;
  int i;

  /* find slot */
  if (!*name)
    return genom_syserr(&(genom_syserr_detail){.code = ENXIO}, &h->context);

  for(i = 0; i < h->ports-><"[$p name]">.n; i++)
    if (!strcmp(h->ports-><"[$p name]">.h[i].name, name)) break;
  if (i >= h->ports-><"[$p name]">.n)
    return genom_syserr(&(genom_syserr_detail){.code = ENXIO}, &h->context);

  p = h->ports-><"[$p name]">.h[i].id;
  if (!p)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  h->ports-><"[$p name]">.h[i].name[0] = 0;
  h->ports-><"[$p name]">.h[i].size = 0;
  h->ports-><"[$p name]">.h[i].id = NULL;
  genom_<"$comp">_client_<"[$p name]">_fini_data(
    <"[[$p datatype] pass reference h->ports->[$p name].h\[i\].data]">);

  if (posterDelete(p) != OK) {
    genom_mwerr_detail d;
    char msg[sizeof(d.what)];
    int err = errnoGet();
    switch(err) {
      case S_posterLib_NOT_OWNER:
      case S_posterLib_POSTER_CLOSED:
        /* ignore those errors, we're fine */ break;

      default:
        h2getErrMsg(err, msg, sizeof(msg));
        if (snprintf(d.what, sizeof(d.what), "%s port: %s", name, msg) < 0)
          strncpy(d.what, "internal error", sizeof(d.what));
        return genom_mwerr(&d, &h->context);
    }
  }
  return genom_ok;
}

<"[[$p datatype] argument reference]">
genom_<"$comp">_client_<"[$p name]">_data(
  genom_client h, const char *name)
{
  int i;

  /* find slot */
  if (!*name) {
    genom_syserr(&(genom_syserr_detail){.code = ENXIO}, &h->context);
    return NULL;
  }

  for(i = 0; i < h->ports-><"[$p name]">.n; i++)
    if (!strcmp(h->ports-><"[$p name]">.h[i].name, name)) break;
  if (i >= h->ports-><"[$p name]">.n) {
    genom_syserr(&(genom_syserr_detail){.code = ENXIO}, &h->context);
    return NULL;
  }

  return <"[[$p datatype] pass reference h->ports->[$p name].h\[i\].data]">;
}

genom_event
genom_<"$comp">_client_<"[$p name]">_write(
  genom_client h, const char *name,
  const <"[[$p datatype] argument reference data]">)
{
  POSTER_ID *p;
  char *addr;
  size_t l;
  int i, s;

  /* find slot */
  if (!*name)
    return genom_syserr(&(genom_syserr_detail){.code = ENXIO}, &h->context);

  for(i = 0; i < h->ports-><"[$p name]">.n; i++)
    if (!strcmp(h->ports-><"[$p name]">.h[i].name, name)) break;
  if (i >= h->ports-><"[$p name]">.n)
    return genom_syserr(&(genom_syserr_detail){.code = ENXIO}, &h->context);

  p = &h->ports-><"[$p name]">.h[i].id;
  if (!*p)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  /* resize if needed */
  l = genom_serialen_<"[[$p datatype] mangle]">(
    <"[[$p datatype] pass value h->ports->[$p name].h\[i\].data]">);
  if (l > h->ports-><"[$p name]">.h[i].size ||
      l + (1<<17) < h->ports-><"[$p name]">.h[i].size) {
    if (posterIoctl(*p, FIO_RESIZE, &l) == OK) {
      h->ports-><"[$p name]">.h[i].size = l;
      goto lock;
    }
    s = errnoGet();
    if (s == S_posterLib_POSTER_CLOSED && posterFind(name, p) == OK) {
      /* ok, found again */
      if (posterIoctl(*p, FIO_RESIZE, &l) == OK) {
        h->ports-><"[$p name]">.h[i].size = l;
        goto lock;
      }
      s = errnoGet();
    }
    goto mwerr;
  }

lock:
  if (posterTake(*p, POSTER_WRITE) != OK) {
    s = errnoGet();
    if (s == S_posterLib_POSTER_CLOSED)
      if (posterFind(name, p) == OK) {
        /* ok, found again */
        if (posterTake(*p, POSTER_WRITE) == OK)
          goto locked;
        s = errnoGet();
      }
    goto mwerr;
  }

locked:
  addr = posterAddr(*p);
  if (data)
    genom_serialize_<"[[$p datatype] mangle]">(
      &addr,
      <"[[$p datatype] pass value [[$p datatype] deref reference data]]">);
  else
    genom_serialize_<"[[$p datatype] mangle]">(
      &addr,
      <"[[$p datatype] pass value h->ports->[$p name].h\[i\].data]">);
  posterGive(*p);
  return genom_ok;

mwerr:
  {
    genom_mwerr_detail d;
    char msg[sizeof(d.what)];

    h2getErrMsg(s, msg, sizeof(msg));
    if (snprintf(d.what, sizeof(d.what), "%s port: %s", name, msg) < 0)
      strncpy(d.what, "internal error", sizeof(d.what));
    return genom_mwerr(&d, &h->context);
  }
}
<'}'>
